__all__ = []

#For each module in this package, import the functions defined in that module
#FIXME:Automate the imports below
from .stanBraMd import *
from .stanKetMd import *
from .indexToKetMd import *
from .schmidtDecompMd import *
from .ketTenQubitMd import *
from .ketTenQubitTwoMd import *

#Add the names of all those functions to the __all__ variable
#FIXME:Automate the additions below
__all__ += stanBraMd.__all__
__all__ += stanKetMd.__all__
__all__ += indexToKetMd.__all__
__all__ += schmidtDecompMd.__all__
__all__ += ketTenQubitMd.__all__
__all__ += ketTenQubitTwoMd.__all__



#isRightDim(mt,da,db)   :   Checks whether mt is square shape with dim = da.db

import numpy as np

__all__ = ['isRightDim']

def isRightDim(mt, da, db):
    r"""Given a numpy.ndarray and two integers da, db returns True if the numpy
    array has shape (d,d) where d = da*db, otherwise returns False.

    Parameters
    ----------
    mt : numpy.ndarray

    da : int

    db : int
    
    Returns
    ----------
    dimCheck : bool
               True if matrix is a da*db square matrix, False otherwise.
    """
    d = da*db
    dimCheck = np.shape(mt) == (d,d)
    return dimCheck    





#indexToKet(i,j,d1,d2)  :   Index (i,j) to ket |ij>

import numpy as np

__all__ = ['indexToKet']

def indexToKet(i,j,d1,d2):
    r""" Takes as input indices (i,j) that represent a ket |ij> over a 
    d1 (tensor) d2 dimensional Hilbert space and returns a 1-d numpy.ndarray 
    representing this ket
    
    Parameters
    ----------
    i  : int 
         state in space 1
    
    j  : int 
         state in space 2
    
    d1 : int 
         dimension of space 1
    
    d2 : int
         dimension of space 2
    
    Returns
    ----------
    vec : numpy.ndarray
          shape (1, d1*d2) vector representing |ij>
    """
    index = d2*i + j
    dim = d1*d2
    vec = np.eye(1,dim,index) 
    
    return vec




#zerOutNum(num)          :   Removes small parts of a number

import copy
import numpy as np

__all__ = ['zerOutNum']

def zerOutNum(num, tol = 1e-15):
    r"""Takes as input a number and tolerance, copies it, in this copy 
    nulls out real and complex parts smaller than the tolerance, and returns
    the copy.
    
    Parameters
    ----------
    num : complex

    tol : float
          Optional, 1e-15 by default
    
    Returns
    ----------
    val : same type as num
    """
    val = copy.deepcopy(num)
    if (np.abs(val.real) < tol):
        val = (val - val.conj())/2.
    if (np.abs(val.imag) < tol):
        val = (val + val.conj())/2.
    return val




#ketTenQubitTwo(i,n)    :   n-qubit state with |0> at i-th tensor and |1> elswhere

import numpy as np

from ..tenProdSelfMd import tenProdSelf
from ..tenProdMd import tenProd

__all__ = ['ketTenQubitTwo']

def ketTenQubitTwo(i,n):
    r"""Takes as input an integer representing a position 'i'
    and another integer representing the number of tensor products 'n'.
    Returns an n-qubit state(numpy.ndarray of length :math:'2^n') with i-th 
    qubit |0> and rest |1>.
    
    Parameters
    ----------
    i : int
        index where ket should be |1>

    n : int
        the number of tensor products
        
    Returns
    ----------
    vec : numpy.ndarray
          A 1-d numpy.ndarray representing an n-qubit state
    
    Notes
    ----------
    The input argument 'ten' should be greater than 1
    """
    
    err1 = 'position of qubit greater or equal to tensor products in ketTenQubit'

    if (i >= n):
        raise ValueError(err1)
    
    zero = np.array([1.,0.])
    one = np.array([0.,1.])
    
    if (i == 0):
        block1 = zero
        block2 = tenProdSelf(one, n-1)

    elif (i == n - 1):
        block1 = tenProdSelf(one, n-1)
        block2 = zero
    
    else:
        block1 = ketTenQubitTwo(i,i+1)
        block2 = tenProdSelf(one,n-i-1)
    
    return tenProd(block1, block2)




#dualMat(mts)           :   Takes list of matrices, computes their dual.

import numpy as np

__all__ = ['dualMat']

def dualMat(mts, lid = False):
    r"""Takes as input a list of matrices, an optional argument stating if
        the matrices are linearly independent and returns a list of matrices
        dual (under Frobenius Norm) to the input list of matrices

    Parameters
    ----------
    mts : list of numpy.ndarrays
          A list of matrices
    
    lid : bool 
          Optional, False by default. If True assumes matrices are 
          linearly independent
    
    Returns
    ----------
    mts : list of numpy.ndarrays
          The dual matrices(under Frobenius Norm) to the set of input matrices. 
          If matrices are linearly independent, it uses regular inverse 
          otherwise uses(less numerically accurate) pseudo-inverse.
    """
    vc = []
    #Each matrix is a ket, place it as a bra making a row of vecRows matrix
    for mt in mts:
        (d,d) = mt.shape
        vcR = np.ndarray.flatten(mt).conj()
        vc += [np.array([vcR])]
    vecRows =  np.vstack(vc)
    #Construct matrix with dual vectors as its columns
    if lid:
        try:
            dualVecCols = np.linalg.inv(vecRows)
        except:
            raise Exception('Cannot find dual matrices')
    else:
        try:
            dualVecCols = np.linalg.pinv(vecRows)
        except:
            raise Exception('Cannot find dual matrices')

    (row, col) = dualVecCols.shape
    mts = []
    #Reshape the columns as matrices of same size as input
    for i in xrange(col):
        mt = dualVecCols[:,i].reshape(d,d)
        mts += [zerOut(mt, tol = 1e-15)]
    return mts




#zerOut(mt)             :   Removes small entries in array

import copy
import numpy as np

__all__ = ['zerOut']

def zerOut(array, tol = 1e-15):
    r"""Takes as input an array and tolerance, copies it, in this copy, 
    nulls out real and complex part of each entry smaller than the tolerance,
    returns the copy.

    Parameters
    ----------
    array : numpy.ndarray

    tol : float
          Optional, 1e-15 by default
    
    Returns
    ----------
    arr : numpy.ndarray
          Identical to input array, except each entry smaller than tol is set
          to zero
    """
    arr = copy.deepcopy(array)
    for index, val in np.ndenumerate(arr):
        if (np.abs(val.real) < tol):
            arr[index] = (arr[index] - arr[index].conj())/2.
        if (np.abs(val.imag) < tol):
            arr[index] = (arr[index] + arr[index].conj())/2.
    return arr




#stanBra(size, index)   :   bra of given 'size' with 1 at index position

import numpy as np

__all__ = ['stanBra']

def stanBra(size, index):
    r"""Takes size and index and returns a row vector of the given size and 1 
    at the given index
    
    Parameters
    ----------
    size : int
    
    index : int
    
    Returns
    ----------
    vec : numpy.ndarray 
          shape (1,size) with 1 at 'index' position 0 elsewhere
    """
    vec = np.eye(1,size,index) 
    return vec




"""
QInfPy: A Python package for Quantum Information
================================================
"""

#Run basic/__init__.py
from .basic import *

#That basic.__init__.py file is written in such a way that running it here 
#will import all functions in the 'basic' sub-package to the namespace of
#this top-level package




#tenProdSelf(mt,k)      :   Tensor product of array with itself k times

from .tenProdMd import tenProd

__all__ = ['tenProdSelf']

def tenProdSelf(mt,k):
    r"""Takes as input a numpy.ndarray mt and an integer k, returns
    the tensor product of the array with itself performed k times
    
    Parameters
    ----------
    mt : numpy.ndarray
    
    k : int
        
    Returns
    ----------
    mtTen : numpy.ndarray
            tensor product of mt with itself k times
    """
    if (k == 1):
        mtTen = mt
    
    else:
        mtTen = tenProd(mt,tenProdSelf(mt,k-1))
 
    return mtTen




#tenProd(v1,v2)         :   Tensor product of two arrays 

import copy

__all__ = ['tenProd']

def tenProd(v1,v2):
    r"""Takes two arrays and returns their tensor product. The function
    just calls the Kronecker product from numpy.

    Parameters
    ----------
    v1 : numpy array
    
    v2 : numpy array
    
    Returns
    ----------
    mat : numpy array
          tensor product of the two inputs
    """
    mat = np.kron(v1,v2) 
    return mat




#getLog(mt)             :   Returns the natural log of the matrix.

import numpy as np
from .isPSDMd import isPSD

__all__ = ['getLog']

def getLog(M, eps=1e-15):
    r"""Takes as input a matrix M and returns the natural log of M.
    
    Parameters
    ----------
    M : numpy.ndarray
        2-d array representing a hermitian matrix
    
    eps : float
          Optional with defaul 1e-15, sets tolerance for the smallest eigenvalue
        
    Returns
    ----------
    lgMt : numpy.ndarray
           log of the input array
    
    Notes
    ----------
    Scales by eps, all eigenvalues between their actual value and 1.0, 
    if any of the eigenvalue is smaller than eps
    """
    try : 
        (psd, val, vec) = isPSD(M,eps,flag=True)
    except :
        raise ValueError('Input matrix is not square and hermitian')

    if psd == False:
        raise ValueError('Eigenvalues of input matrix not sufficiently positive')
    
    n = len(val)
    
    #If any of the eigenvalues is smaller than eps, then rescale the spectrum
    #to make all eigenvalues at least eps, this prevents log from complaining
    
    if np.any(val<eps):
        val = (1-eps)*val + eps*1. 
    
    lgMt = np.dot(np.log(val)*vec,vec.conj().T)
    return lgMt




#doLinComb(...)         :   Takes array of numbers and matrices, constructs linear combination

import numpy as np

__all__ = ['doLinComb']

def doLinComb(num, mats):
    r"""Takes in a 1-d numpy.ndarray of numbers, and a list of matrices
    and returns a linear combination of the matrices.
    
    Parameters
    ----------
    num : numpy.ndarray

    mats : A list of numpy.ndarrays

    Returns
    ----------
    rho : numpy.ndarray
          Multiplies each number in num with corresponding matrix in mats and
          constructs a linear combination
    """
    if (len(num) != len(mats)):
        raise ValueError('Length of number vector and number of matrices unequal')
        
    (da, da) = mats[0].shape
    rho = 1j*np.zeros(shape = (da, da))
    for i in xrange(len(num)):
        rho += mats[i]*num[i]
    return rho




#schmidtDecomp(...)   :   Gives Schmidt decomposition over spaces of given dimension

import numpy as np

__all__ = ['schmidtDecomp']

def schmidtDecomp(psi,da,db):
    r"""Takes an input a vector, and dimensions :math:'da' and :math:'db'
    of the two spaces over whose tensor product the vector is defined.
    Returns the Schmidt decomposition of the vector.
   
    Parameters
    ----------
    psi : numpy.ndarray
          1-d complex vector
    
    da : int
         Dimension of space 1

    db : int
         Dimension of space 2
        
    Returns
    ----------
    comp : numpy.ndarray
           A vector with components that are the Schmidt coefficients of the 
           input psi

    U : numpy.ndarray
        Matrix with :math:'k^{th}' Schmidt vector on space 1 written as U[:,k]

    V : numpy.ndarray
        Matrix with :math:'k^{th}' Schmidt vector on space 2 written as V[:,k]
    """
    try:
        mat = psi.reshape(da,db)
    except:
        raise ValueError('Shape of psi and da*db do not seem compatible')
        return None
    U, D, V = np.linalg.svd(mat, full_matrices=True)
    return (D,U,V.T)




#hermBasis(n)           :   Takes integer n, returns n^2-1 traceless, Hermitian basis

import numpy as np

__all__ = ['hermBasis']

def __getW(j,n):
    r"""Takes as input 2 integers j, n (expects 1<= j <= n-1) and returns
    a diagonal operator constructed recursively. 
    Helper Function for hermBasis(n)
    
    Parameters
    ----------
    j : int
        index for diagonal operator
    
    n : int
        dimension of diagonal operator

    Returns
    ----------
    returns : numpy.ndarray
              A diagonal matrix operator, constructed recursively. It is the 
              w matrix of Sec. 2.3 in the 'Generalized Bloch Vector and the 
              Eigenvalues of a Density Matrix' paper by Ozols and Mancinska
    """
    if (j >= n-1):
        raise ValueError('Invalid indices')
    
    if (j == 0 ):
        mat = np.zeros(shape=(n,n))
        mat[0,0] = 1.
        mat[1,1] = -1.
        return mat
    else:
        mat = np.zeros(shape=(n,n))
        mat[j,j] = j+1.
        mat[j+1,j+1] = -(j+1.)
        m1 = __getW(j-1, n)*np.sqrt(j*(j+1)/2.)
        return (m1 + mat)*np.sqrt(2./((j+1.)*(j+2.))) 


def hermBasis(n):
    r"""Takes as input a integer n and generates n^2 - 1, traceless,
    hermitian, orthogonal, norm 2 matrices 
    
    Parameters
    ----------
    n : int
        Dimension of space

    Returns
    ----------
    basis : list of numpy.ndarrays
            A list of length n^2 - 1, traceless, hermitian, orthogonal, norm 2 
            basis operators. They are the :math:'\lambda' matrix of Sec. 2.3 
            in the 'Generalized Bloch Vector and the Eigenvalues of a 
            Density Matrix' paper by Ozols and Mancinska
    """
    basis = []
    for i in xrange(0, n-1):
        basis += [__getW(i,n)]
    for j in xrange(0,n-1):
        for k in xrange(j+1,n):
            mat = np.zeros(shape=(n,n))
            mat[k,j] = 1.
            u = mat + mat.T
            v = 1j*(mat - mat.T)
            basis += [u]
            basis += [v]
    return basis



#isPSD(mt)              :   Checks whether matrix is PSD or not

import numpy as np

__all__ = ['isPSD']

def isPSD(mt, tol = 1e-12, flag = False):
    r"""Takes a matrix, returns True if is Positive Semi-definite, otherwise
    returns False

    Parameters
    ----------
    mt : numpy.ndarray

    tol : float
          Optional, default value 1e-12

    flag : bool
           Optional, Returns eigenvalues and of mt if True

    Returns
    ----------
    When flag is False, returns pos, else returns (pos, val, vec)

    pos : bool
          True if mt is positive semi-definite up to some numerical tolerance; 
          False otherwise

    val : numpy.ndarray
          Eigenvalues of mt

    vec : numpy.ndarray
          Eigenvectors of mt

    """
    #Square Matrix
    (row, col) = np.shape(mt)
    sqr = row == col
    if not sqr:
        return False

    #Hermitian
    mtDag = mt.conj().T
    diff = mt - mtDag
    herm = np.linalg.norm(diff) < 10*tol
    if not herm:
        return False

    #Positivity
    mt2 = (mt + mt.conj().T)/2.
    val, vec = np.linalg.eigh(mt2)
    pos = all(val.real > -tol) & all(val.imag > -tol)
    
    if flag:
        return (pos, val, vec)
    else:
        return pos




__all__ = []

#For each module in this package, import the functions defined in that module
#FIXME: Automate the imports below
from .doLinCombMd import *
from .dualMatMd import *
from .getLogMd import *
from .hermBasisMd import *
from .isPSDMd import *
from .isRightDimMd import *
from .tenProdMd import *
from .tenProdSelfMd import *
from .zerOutNumMd import *
from .zerOutMd import *

#Add the names of all those functions to the __all__ variable
#FIXME: Automate the additions below
__all__ += doLinCombMd.__all__
__all__ += dualMatMd.__all__
__all__ += getLogMd.__all__
__all__ += hermBasisMd.__all__
__all__ += isPSDMd.__all__
__all__ += isRightDimMd.__all__
__all__ += tenProdMd.__all__
__all__ += tenProdSelfMd.__all__
__all__ += zerOutNumMd.__all__
__all__ += zerOutMd.__all__


#Run basicOne/__init__.py. The way that __init__.py file is written, we import 
#all functions in that sub-package
from . import pureState
from .pureState import *

__all__ += pureState.__all__

